//
// Copyright (c) 2013-2016 Pavel Medvedev. All rights reserved.
//
// This file is part of v8pp (https://github.com/pmed/v8pp) project.
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//
#include "v8pp/convert.hpp"
#include "test.hpp"

#include <list>
#include <vector>
#include <map>

template<typename T>
void test_conv(v8::Isolate* isolate, T value)
{
	v8::Local<v8::Value> v8_value = v8pp::to_v8(isolate, value);
	auto const value2 = v8pp::from_v8<T>(isolate, v8_value);
	check_eq(typeid(T).name(), value2, value);
}

template<typename Char, size_t N>
void test_string_conv(v8::Isolate* isolate, Char const (&str)[N])
{
	std::basic_string<Char> const str2(str, 2);

	test_conv(isolate, str[0]);
	test_conv(isolate, str);

	check_eq("string literal",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, str)), str);
	check_eq("string literal2",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, str, 2)), str2);

	Char const* ptr = str;
	check_eq("string pointer",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, ptr)), str);
	check_eq("string pointer2",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, ptr, 2)), str2);

	Char const* empty = str + N - 1; // use last \0 in source string
	check_eq("empty string literal",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, "")), empty);
	check_eq("empty string literal0",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, "", 0)), empty);

	check_eq("empty string pointer",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, empty)), empty);
	check_eq("empty string pointer0",
		v8pp::from_v8<Char const*>(isolate, v8pp::to_v8(isolate, empty, 0)), empty);
}

struct person
{
	std::string name;
	int age;

//for test framework
	bool operator!=(person const& other) const
	{
		return name != other.name || age != other.age;
	}

	friend std::ostream& operator<<(std::ostream& os, person const& p)
	{
		return os << "person: " << p.name << " age: " << p.age;
	}
};

namespace v8pp {

template<>
struct convert<person>
{
	using from_type = person;
	using to_type = v8::Handle<v8::Object>;

	static bool is_valid(v8::Isolate*, v8::Handle<v8::Value> value)
	{
		return !value.IsEmpty() && value->IsObject();
	}

	static to_type to_v8(v8::Isolate* isolate, person const& p)
	{
		v8::EscapableHandleScope scope(isolate);
		v8::Local<v8::Object> obj = v8::Object::New(isolate);
		obj->Set(v8pp::to_v8(isolate, "name"), v8pp::to_v8(isolate, p.name));
		obj->Set(v8pp::to_v8(isolate, "age"), v8pp::to_v8(isolate, p.age));
		/* Simpler after #include <v8pp/object.hpp>
		set_option(isolate, obj, "name", p.name);
		set_option(isolate, obj, "age", p.age);
		*/
		return scope.Escape(obj);
	}

	static from_type from_v8(v8::Isolate* isolate, v8::Handle<v8::Value> value)
	{
		if (!is_valid(isolate, value))
		{
			throw std::runtime_error("excpected object");
		}

		v8::HandleScope scope(isolate);
		v8::Local<v8::Object> obj = value.As<v8::Object>();

		person result;
		result.name = v8pp::from_v8<std::string>(isolate,
			obj->Get(v8pp::to_v8(isolate, "name")));
		result.age = v8pp::from_v8<int>(isolate,
			obj->Get(v8pp::to_v8(isolate, "age")));
		
		/* Simpler after #include <v8pp/object.hpp>
		get_option(isolate, obj, "name", result.name);
		get_option(isolate, obj, "age", result.age);
		*/
		return result;
	}
};

} // v8pp

void test_convert()
{
	v8pp::context context;
	v8::Isolate* isolate = context.isolate();
	v8::HandleScope scope(isolate);

	test_conv(isolate, 1);
	test_conv(isolate, 2.2);
	test_conv(isolate, true);

	enum old_enum { A = 1, B = 5, C = - 1 };
	test_conv(isolate, B);

	enum class new_enum { X = 'a', Y = 'b', Z = 'c' };
	test_conv(isolate, new_enum::Z);

	test_string_conv(isolate, "qaz");
#ifdef WIN32
	test_string_conv(isolate, L"qaz");
#endif

	using int_vector = std::vector<int>;

	int_vector vector = { 1, 2, 3 };
	test_conv(isolate, vector);

	std::map<char, int> map = { { 'a', 1 }, { 'b', 2 }, { 'c', 3 } };
	test_conv(isolate, map);

	std::array<int, 3> array = { 1, 2, 3 };
	test_conv(isolate, array);

	check_ex<std::runtime_error>("wrong array length", [isolate, array]()
	{
		v8::Local<v8::Array> arr = v8pp::to_v8(isolate,
			std::array<int, 3>{ 1, 2, 3 });
		v8pp::from_v8<std::array<int, 2>>(isolate, arr);
	});

	check_eq("initializer list to array",
		v8pp::from_v8<int_vector>(isolate, v8pp::to_v8(isolate,
		{ 1, 2, 3 })), vector);

	std::list<int> list = { 1, 2, 3 };
	check_eq("pair of iterators to array", v8pp::from_v8<int_vector>(isolate,
		v8pp::to_v8(isolate, list.begin(), list.end())), vector);

	person p;
	p.name = "Al"; p.age = 33;
	test_conv(isolate, p);
}
